from __future__ import absolute_import

from symstore.symstore import Store
from symstore.symstore import Transactions
from symstore.symstore import History
from symstore.symstore import Transaction
from symstore.symstore import TransactionEntry
from symstore.version import VERSION
from symstore.errs import FileFormatError


__all__ = ["Store", "Transactions", "History", "Transaction",
           "TransactionEntry", "FileFormatError", "VERSION"]



from __future__ import absolute_import

import os
import io
import struct
from symstore import errs

PE_SIGNATURE = b"PE\0\0"
PE_SIGNATURE_POINTER = 0x3C

PE_SIG_SIZE = 4
MACHINE_SIZE = 2
NUMBER_OF_SECTION_SIZE = 2
TIME_DATE_STAMP_SIZE = 4
POINTER_TO_SYMBOL_TABLE_SIZE = 4
NUMBER_OF_SYMBOLS_SIZE = 4
SIZE_OF_OPTIONAL_HEADER_SIZE = 2
CHARACTERISTICS_SIZE = 2

# TimeDateStamp field's offset relative to PE signature
TIME_DATE_STAMP_OFFSET = \
    PE_SIG_SIZE + \
    MACHINE_SIZE + \
    NUMBER_OF_SECTION_SIZE

# Optional header offset relative to PE signature
OPTIONAL_HEADER_OFFSET = \
    PE_SIG_SIZE + \
    MACHINE_SIZE + \
    NUMBER_OF_SECTION_SIZE + \
    TIME_DATE_STAMP_SIZE + \
    POINTER_TO_SYMBOL_TABLE_SIZE + \
    NUMBER_OF_SYMBOLS_SIZE + \
    SIZE_OF_OPTIONAL_HEADER_SIZE + \
    CHARACTERISTICS_SIZE

# SizeOfImage field's offset relative to optional header start
SIZE_OF_IMAGE_OFFSET = 56


class PEFormatError(errs.FileFormatError):
    format_name = "PE"


def _read_i32(file, size, offset):
    """
    Read 32-bit little-endian integer from an opened file.

    :param file: opended file handle
    :param size: file szie
    :param offset: the offset of the integer
    """
    if offset + 4 > size:
        raise PEFormatError("data offset %s beyond end of file %s" %
                            (offset, size))

    file.seek(offset)

    return struct.unpack("<i", file.read(4))[0]


class PEFile:
    """
    Simple PE file parser, that loads two fields used by symstore:

    * TimeDateStamp from file header
    * SizeOfImage from optional header

    The values are accessed by reading the object's member variables by
    the same name, e.g.

        PEFile("some.exe").TimeDateStamp
    """
    def __init__(self, filepath):
        with io.open(filepath, "rb") as f:
            fsize = os.fstat(f.fileno()).st_size

            # load PE signature offset
            pe_sig_offset = _read_i32(f, fsize, PE_SIGNATURE_POINTER)

            # check that file contains valid PE signature
            f.seek(pe_sig_offset)
            if f.read(4) != PE_SIGNATURE:
                raise PEFormatError("PE signature not found")

            # load TimeDateStamp field
            self.TimeDateStamp = \
                _read_i32(f, fsize, pe_sig_offset+TIME_DATE_STAMP_OFFSET)

            # load SizeOfImage field
            self.SizeOfImage = _read_i32(f, fsize,
                                         pe_sig_offset +
                                         OPTIONAL_HEADER_OFFSET +
                                         SIZE_OF_IMAGE_OFFSET)



from __future__ import absolute_import

import os
import re
import time
import shutil
from os import path
from symstore import pe
from symstore import pdb
from datetime import datetime

TRANSACTION_LINE_RE = re.compile(
    r"(\d+),"
    r"(add|del),"
    r"(file|ptr),"
    r"((?:\d\d/\d\d/\d\d\d\d),(?:\d\d:\d\d:\d\d)),"
    "\"([^\"]*)\","
    "\"([^\"]*)\","
    "\"([^\"]*)\","
    r".*")

ADMIN_DIR = "000Admin"
LAST_ID_FILE = path.join(ADMIN_DIR, "lastid.txt")
HISTORY_FILE = path.join(ADMIN_DIR, "history.txt")
SERVER_FILE = path.join(ADMIN_DIR, "server.txt")
PINGME_FILE = "pingme.txt"

PDB_IMAGE, PE_IMAGE = range(2)

EXT_TYPES = dict(pdb=PDB_IMAGE,
                 exe=PE_IMAGE,
                 dll=PE_IMAGE)


def _pdb_hash(filename):
    pdbfile = pdb.PDBFile(filename)
    return "%s%s" % (pdbfile.guid, pdbfile.age)


def _pe_hash(file):
    pefile = pe.PEFile(file)

    return "%X%X" % (pefile.TimeDateStamp, pefile.SizeOfImage)


def _image_type(file):
    file_ext = path.splitext(file)[1][1:].lower()
    # TODO handle cases of unknown file extensions
    return EXT_TYPES[file_ext]


def _file_hash(file):

    image_type = _image_type(file)

    if image_type == PDB_IMAGE:
        return _pdb_hash(file)

    assert image_type == PE_IMAGE
    return _pe_hash(file)


class TransactionEntry:
    def __init__(self, symstore, file_name, file_hash, source_file):
        self._symstore = symstore
        self.file_name = file_name
        self.file_hash = file_hash
        self.source_file = source_file

    def open(self):
        fpath = path.join(self._symstore._path, self.file_name,
                          self.file_hash, self.file_name)

        return open(fpath, "rb")

    def publish(self):
        """
        publish this entry's source file inside symstore
        """
        dest_dir = path.join(self._symstore._path,
                             self.file_name,
                             self.file_hash)
        os.makedirs(dest_dir)
        shutil.copy(self.source_file, dest_dir)
        # TODO handle I/O errors

    def __str__(self):
        return """"%s\%s","%s""""" % \
               (self.file_name, self.file_hash,
                path.abspath(self.source_file))


class Transaction:
    transaction_entry_class = TransactionEntry

    def __init__(self, symstore, id=None, type="add", ref="file",
                 timestamp=None, product=None, version=None, comment=None):

        self._symstore = symstore
        self._entries = None
        self.id = id
        self.type = type
        self.ref = ref
        self.timestamp = timestamp
        self.product = product
        self.version = version
        self.comment = comment

    def _commited(self):
        return self.id is not None

    def _entries_file(self, mode="r"):
        return open(path.join(self._symstore._admin_dir, self.id),
                    mode=mode)

    def _load_entries(self):
        if not self._commited():
            return []

        entries = []
        with self._entries_file() as efile:
            for line in efile.readlines():
                entry, source_file = [
                    s.strip("\"") for s in line.strip().split(",")]

                file_name, file_hash = entry.split("\\")

                transaction_entry = self.transaction_entry_class(
                    self._symstore, file_name, file_hash, source_file)

                entries.append(transaction_entry)

        # TODO catch IOERrror
        # TODO catch parse errors

        return entries

    def add_file(self, file):
        """
        :raises pe.PEFormatError: on errors parsing PE (.exe/.dll) files
        """
        entry = TransactionEntry(self._symstore,
                                 path.basename(file),
                                 _file_hash(file),
                                 file)
        # TODO handle I/O errors from _file_hash()

        self.entries.append(entry)

    @property
    def entries(self):
        if self._entries is None:
            self._entries = self._load_entries()

        return self._entries

    def commit(self, id, now):
        assert not self._commited()

        self.timestamp = now
        self.id = id

        # publish all entries files to the store
        for entry in self.entries:
            entry.publish()

        # write new transaction file
        with self._entries_file("a") as efile:
            for entry in self.entries:
                efile.write("%s\n" % entry)
        # TODO handle I/O errors while opening/writing efile

    def __str__(self):
        date_stamp = self.timestamp.strftime("%m/%d/%Y")
        time_stamp = self.timestamp.strftime("%H:%M:%S")

        assert self._commited()
        return """%s,%s,%s,%s,%s,"%s","%s","",""" % \
               (self.id, self.type, self.ref, date_stamp, time_stamp,
                self.product, self.version)


def parse_transaction_line(line):
    # TODO handle parse errors in this function
    (id, type, ref, timestamp, product, version, comment) = \
        TRANSACTION_LINE_RE.match(line).groups()

    ts = datetime.strptime(timestamp, "%m/%d/%Y,%H:%M:%S")
    return id, type, ref,  ts, product, version, comment


class Transactions:
    transaction_class = Transaction

    def __init__(self, symstore):
        self._symstore = symstore
        self._transactions = None

    def _server_file(self, mode="r"):
        return open(self._symstore._server_file, mode=mode)

    def _server_file_exists(self):
        return path.isfile(self._symstore._server_file)

    def _parse_server_file(self):
        if not self._server_file_exists():
            return {}

        transactions = {}

        with self._server_file() as sfile:
            for line in sfile.readlines():
                transaction = self.transaction_class(
                    self._symstore, *parse_transaction_line(line))

                transactions[transaction.id] = transaction

        return transactions

    def _get_transactions(self):
        if self._transactions is None:
            self._transactions = self._parse_server_file()
        return self._transactions

    def items(self):
        return self._get_transactions().items()

    def add(self, transaction):
        with self._server_file("a") as sfile:
            sfile.write("%s\n" % transaction)
        # TODO handle I/O errors


class History:
    transaction_class = Transaction

    def __init__(self, symstore):
        self._symstore = symstore
        self._transactions = None

    def _history_file(self, mode="r"):
        return open(self._symstore._history_file, mode=mode)

    def _history_file_exists(self):
        return path.isfile(self._symstore._history_file)

    def _parse_history_file(self):
        if not self._history_file_exists():
            return []

        transactions = []

        with self._history_file() as hfile:
            for line in hfile.readlines():
                transaction = self.transaction_class(
                    self._symstore, *parse_transaction_line(line))

                transactions.append(transaction)

        return transactions

    def _get_transactions(self):
        if self._transactions is None:
            self._transactions = self._parse_history_file()
        return self._transactions

    def __len__(self):
        return len(self._get_transactions())

    def __getitem__(self, item):
        return self._get_transactions()[item]

    def add(self, transaction):
        with self._history_file("a") as hfile:
            if hfile.tell() != 0:
                # add line break if appending to existing non-empty file
                hfile.write("\n")
            hfile.write("%s" % transaction)

        # TODO handle I/O errors


class Store:
    def __init__(self, store_path):
        self._path = store_path
        self.transactions = Transactions(self)
        self.history = History(self)

    @property
    def modify_timestamp(self):
        """
        Get the time when this symstore was last modified.

        The modify timestamp is returned as datetime.datetime object.
        """
        return datetime.fromtimestamp(os.stat(self._pingme_file).st_mtime)

    @property
    def _admin_dir(self):
        return path.join(self._path, ADMIN_DIR)

    @property
    def _last_id_file(self):
        return path.join(self._path, LAST_ID_FILE)

    @property
    def _history_file(self):
        return path.join(self._path, HISTORY_FILE)

    @property
    def _server_file(self):
        return path.join(self._path, SERVER_FILE)

    @property
    def _pingme_file(self):
        return path.join(self._path, PINGME_FILE)

    def _create_dirs(self):
        if not path.isdir(self._path):
            os.mkdir(self._path)
            # TODO handle mkdir errors

        admin_dir = self._admin_dir
        if not path.isdir(admin_dir):
            os.mkdir(admin_dir)
            # TODO handle mkdir errors

    def _next_transaction_id(self):
        last_id_file = self._last_id_file

        last_id = 0
        if path.isfile(last_id_file):
            # TODO handle open and read errors
            # TODO handle parse errors
            last_id = int(open(last_id_file, "r").read())

        return "%.010d" % (last_id + 1)

    def _write_transaction_id(self, trans_id):
        with open(self._last_id_file, "w") as id_file:
            id_file.write(trans_id)

    def _touch_pingme(self, timestamp):
        pingme_path = self._pingme_file

        if not path.isfile(pingme_path):
            open(pingme_path, "a")

        os.utime(pingme_path, (timestamp, timestamp))

    def new_transaction(self, product, version, type="add"):
        return Transaction(self, type=type, product=product, version=version)

    def commit(self, transaction):
        self._create_dirs()

        now = round(time.time())

        transaction.commit(self._next_transaction_id(),
                           datetime.fromtimestamp(now))

        self.transactions.add(transaction)
        self.history.add(transaction)

        self._write_transaction_id(transaction.id)
        self._touch_pingme(now)



#!/usr/bin/env python
import setuptools
from os import path

VERSION = open(path.join("symstore", "VERSION")).read().strip()

setuptools.setup(
    name="symstore",
    version=VERSION,
    description="publish PDB and PE files to symbols store",
    long_description=open("README.rst", "r").read(),
    url="https://github.com/elmirjagudin/symstore",
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
    ],
    keywords="development symstore pdb",
    packages=["symstore"],
    package_data={"symstore": ["VERSION"]},
    scripts=["symstore/bin/symstore"],
    extras_require={
        "develop":
            [
                "coverage",
                "flake8",
                "mock"
            ]
    }
)



from __future__ import absolute_import

import io
import math
import struct
import binascii
from symstore import errs

SIGNATURE = b"Microsoft C/C++ MSF 7.00\r\n\x1ADS\0\0\0"


class PDBFormatError(errs.FileFormatError):
    format_name = "PDB"


def pages(size, page_size):
    """
    calculate number of pages that are required to store the specified
    number of bytes

    :param size: number of bytes
    :param page_size: page size
    """
    return int(math.ceil(float(size)/page_size))


class Root:
    """
    A bare bones abstraction of the root stream of an PDB files. Provides
    methods to read raw chunks of the root stream, as well as accessing
    data about streams containing streams.
    """
    def __init__(self, fp, page_size, size, ptr):
        self.fp = fp
        self.page_size = page_size
        self.size = size
        self.ptr = ptr

    def _pages(self):
        """
        get page numbers used by the root stream
        """
        self.fp.seek(self.ptr * self.page_size)
        num_pages = pages(self.size, self.page_size)
        assert num_pages * 4 < self.page_size, "long root stream not supported"
        return struct.unpack("<%sI" % num_pages,
                             self.fp.read(4*num_pages))

    def _seek(self, page, byte):
        """
        move the root stream to a new position

        :param page: move to this page
        :param byte: move to the offset in the specified page
        """
        offset = self._pages()[page] * self.page_size + byte
        self.fp.seek(offset)

    def read(self, start, length):
        """
        read root stream bytes

        :param start: the global offset where to read
        :param length: number of bytes to read

        :return: root bytes as an bytes array
        """
        assert length < self.page_size, \
            "multi-page reads not supported"

        start_page = start // self.page_size
        start_byte = start % self.page_size

        end = start + length - 1
        end_page = end // self.page_size

        assert start_page == end_page, \
            "reads spanning two pages not implemented"

        self._seek(start_page, start_byte)
        return self.fp.read(length)

    def num_streams(self):
        """
        get number of streams listed in this root stream
        :return:
        """
        return struct.unpack("<I", self.read(0, 4))[0]

    def stream_size(self, stream_index):
        """
        get stream's size in bytes
        """
        if stream_index >= self.num_streams():
            raise IndexError("stream index to large")

        return struct.unpack("<I", self.read(stream_index * 4 + 4, 4))[0]

    def stream_pages(self, stream_index):
        """
        get stream's page numbers
        """
        num_streams = self.num_streams()

        pages_offset = 4 + 4 * num_streams

        for sidx in range(stream_index):
            stream_size = self.stream_size(sidx)
            stream_num_pages = pages(stream_size, self.page_size)
            pages_offset += stream_num_pages * 4

        num_pages = pages(self.stream_size(stream_index),
                          self.page_size)

        return struct.unpack("<%dI" % num_pages,
                             self.read(pages_offset, 4*num_pages))


class GUID:
    """
    Represents Globally Unique Identifier (GUID). The value is accessable
    via 4 data fields with sizes below:

        data1 - 32bit integer
        data2 - 16bit integer
        data3 - 16bit integer
        data4 - 64bit as byte array

    This object can also be converted to a hexidecimal string representation.
    """
    def __init__(self, data1, data2, data3, data4):
        self.data1 = data1
        self.data2 = data2
        self.data3 = data3
        self.data4 = data4

    def __str__(self):
        data4_str = binascii.hexlify(self.data4).decode("utf-8").upper()

        return "%.8X%.4X%.4X%s" % \
               (self.data1, self.data2, self.data3, data4_str)


class PDBFile:
    """
    Simple PDB (program database) file parser the loads files GUID and Age
    fields.

    The GUID and Age values are accessable as following object members:

        guid - PDB file's GUID as an instance of GUID class
        age  - PDB file's Age, as an integer
    """
    def __init__(self, filepath):
        with io.open(filepath, "rb") as f:

            # Check signature
            sig = f.read(len(SIGNATURE))
            if sig != SIGNATURE:
                raise PDBFormatError("Invalid signature")

            # load page size and root stream definition
            page_size, _, _, root_dir_size, _, root_ptr = \
                struct.unpack("<IIIIII", f.read(4*6))

            # Create Root stream parser
            root = Root(f, page_size, root_dir_size, root_ptr)

            # load the PDB stream page
            pdb_stream_pages = root.stream_pages(1)
            f.seek(pdb_stream_pages[0]*page_size)

            # load age and GUID from PDB stream
            _, _, age, guid_d1, guid_d2, guid_d3, guid_d4 = \
                struct.unpack("<IIIIHH8s", f.read(4*4 + 2 * 2 + 8))

            # store GUID and age for user friendly retrieval
            self.guid = GUID(guid_d1, guid_d2, guid_d3, guid_d4)
            self.age = age
